/************************************************************************************
*                                                                                   *
*   Copyright (c) 2013 Axel Menzel <info@axelmenzel.de>                             *
*                                                                                   *
*   This file is part of the Runtime Type Reflection System (RTTR).                 *
*                                                                                   *
*   Permission is hereby granted, free of charge, to any person obtaining           *
*   a copy of this software and associated documentation files (the "Software"),    *
*   to deal in the Software without restriction, including without limitation       *
*   the rights to use, copy, modify, merge, publish, distribute, sublicense,        *
*   and/or sell copies of the Software, and to permit persons to whom the           *
*   Software is furnished to do so, subject to the following conditions:            *
*                                                                                   *
*   The above copyright notice and this permission notice shall be included in      *
*   all copies or substantial portions of the Software.                             *
*                                                                                   *
*   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR      *
*   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,        *
*   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE     *
*   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER          *
*   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,   *
*   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE   *
*   SOFTWARE.                                                                       *
*                                                                                   *
*************************************************************************************/

#define BOOST_ERROR_CODE_HEADER_ONLY
#define BOOST_CHRONO_HEADER_ONLY
#define BOOST_SYSTEM_NO_LIB

#include <boost/chrono.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/make_shared.hpp>


#include "TestClasses.h"

#define ITEM_COUNT 50000
#define RUN_COUNT 5
#include "BoostTest.h"

#include <iostream>

using namespace RTTR;

struct BenchmarkResult
{
    BenchmarkResult(const std::string& title, boost::chrono::process_real_cpu_clock::duration time) : _title(title), _timing(time) {}
    std::string                                     _title;
    boost::chrono::process_real_cpu_clock::duration _timing;
};

void printStats(const std::vector<BenchmarkResult>& timingResults)
{
    std::cout << "-------------------------------------------------------\n";
    for (std::vector<BenchmarkResult>::const_iterator itr = timingResults.begin();
         itr != timingResults.end();
         ++itr)
    {
        if (!(*itr)._title.empty())
            std::cout << (*itr)._title << ": " << boost::chrono::duration_cast<boost::chrono::milliseconds>((*itr)._timing).count() << " ms\n";
    }
    std::cout << "-------------------------------------------------------\n";
}

template<typename T>
static std::vector<T*> transform(std::vector<boost::shared_ptr<T> >& vec)
{
    std::vector<T*> result;
    result.reserve(vec.size());
    for(typename std::vector<boost::shared_ptr<T> >::const_iterator itr = vec.begin();
        itr != vec.end();
        ++itr)
    {
        result.push_back((*itr).get());
    }
    return result;
}

BOOST_AUTO_TEST_SUITE(PerformanceTests)

BOOST_AUTO_TEST_CASE(init_PerformanceTest)
{
    std::cout << "Benchmark: dynamic_cast vs. rttr_cast / typeid vs. TypeInfo::get" << std::endl;
}

BOOST_AUTO_TEST_CASE(testSingleClassHierachy_PerformanceTest)
{
    using namespace boost::chrono;
    std::vector<BenchmarkResult> timingResults;
    std::vector<boost::shared_ptr<ClassSingleBase> > shared_vec;
    shared_vec.reserve(ITEM_COUNT * 5);
    for (int i = 0; i < ITEM_COUNT; ++i)
    {
        shared_vec.push_back(boost::make_shared<ClassSingle1A>());
        shared_vec.push_back(boost::make_shared<ClassSingle1B>());
        shared_vec.push_back(boost::make_shared<ClassSingle1C>());
        shared_vec.push_back(boost::make_shared<ClassSingle1D>());
        shared_vec.push_back(boost::make_shared<ClassSingle1E>());
    }

    std::vector<ClassSingleBase*> vec = transform(shared_vec);
    
    process_real_cpu_clock::time_point startTime;
    process_real_cpu_clock::duration elapsedTime;
    std::vector<int> result;
    result.resize(ITEM_COUNT * 5);

    for (int count = 0; count < RUN_COUNT; ++count)
    {
        startTime = process_real_cpu_clock::now();
        for (int i = 0; i < ITEM_COUNT * 5; ++i)
        {
            if (ClassSingle1A* item = dynamic_cast<ClassSingle1A*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle1B* item = dynamic_cast<ClassSingle1B*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle1C* item = dynamic_cast<ClassSingle1C*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle1D* item = dynamic_cast<ClassSingle1D*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle1E* item = dynamic_cast<ClassSingle1E*>(vec[i]))
            {
                result[i] = item->getType();
            }
        }
        elapsedTime += process_real_cpu_clock::now() - startTime;
    }
    elapsedTime /= RUN_COUNT;
    timingResults.push_back(BenchmarkResult("ClassSingleBase to Level 1 [dynamic_cast]      ", elapsedTime));
    elapsedTime = process_real_cpu_clock::duration();

    for (int count = 0; count < RUN_COUNT; ++count)
    {
        startTime = process_real_cpu_clock::now();
        for (int i = 0; i < ITEM_COUNT * 5; ++i)
        {
            if (ClassSingle1A* item = rttr_cast<ClassSingle1A*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle1B* item = rttr_cast<ClassSingle1B*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle1C* item = rttr_cast<ClassSingle1C*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle1D* item = rttr_cast<ClassSingle1D*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle1E* item = rttr_cast<ClassSingle1E*>(vec[i]))
            {
                result[i] = item->getType();
            }
        }
        elapsedTime += process_real_cpu_clock::now() - startTime;
    }
 
    elapsedTime /= RUN_COUNT;
    timingResults.push_back(BenchmarkResult("ClassSingleBase to Level 1 [rttr_cast]         ", elapsedTime));
    printStats(timingResults);
}

/////////////////////////////////////////////////////////////////////////////////////////

BOOST_AUTO_TEST_CASE(testSingleClassHierachyMiddle_PerformanceTest)
{
    using namespace boost::chrono;
    std::vector<BenchmarkResult> timingResults;
    std::vector<boost::shared_ptr<ClassSingleBase> > shared_vec;

    shared_vec.reserve(ITEM_COUNT * 5);
    for (int i = 0; i < ITEM_COUNT; ++i)
    {
        shared_vec.push_back(boost::make_shared<ClassSingle3A>());
        shared_vec.push_back(boost::make_shared<ClassSingle3B>());
        shared_vec.push_back(boost::make_shared<ClassSingle3C>());
        shared_vec.push_back(boost::make_shared<ClassSingle3D>());
        shared_vec.push_back(boost::make_shared<ClassSingle3E>());
    }
    std::vector<ClassSingleBase*> vec = transform(shared_vec);

    process_real_cpu_clock::time_point startTime;
    process_real_cpu_clock::duration elapsedTime;

    std::vector<int> result;
    result.resize(ITEM_COUNT * 5);
    for (int count = 0; count < RUN_COUNT; ++count)
    {
        startTime = process_real_cpu_clock::now();
        for (int i = 0; i < ITEM_COUNT * 5; ++i)
        {
            if (ClassSingle3A* item = dynamic_cast<ClassSingle3A*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle3B* item = dynamic_cast<ClassSingle3B*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle3C* item = dynamic_cast<ClassSingle3C*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle3D* item = dynamic_cast<ClassSingle3D*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle3E* item = dynamic_cast<ClassSingle3E*>(vec[i]))
            {
                result[i] = item->getType();
            }
        }
        elapsedTime += process_real_cpu_clock::now() - startTime;
    }
    elapsedTime /= RUN_COUNT;
    timingResults.push_back(BenchmarkResult("ClassSingleBase to Level 3 [dynamic_cast]      ", elapsedTime));
    
    elapsedTime = process_real_cpu_clock::duration();
    for (int count = 0; count < RUN_COUNT; ++count)
    {
        startTime = process_real_cpu_clock::now();

        for (int i = 0; i < ITEM_COUNT * 5; ++i)
        {
            if (ClassSingle3A* item = rttr_cast<ClassSingle3A*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle3B* item = rttr_cast<ClassSingle3B*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle3C* item = rttr_cast<ClassSingle3C*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle3D* item = rttr_cast<ClassSingle3D*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle3E* item = rttr_cast<ClassSingle3E*>(vec[i]))
            {
                result[i] = item->getType();
            }
        }
        elapsedTime += process_real_cpu_clock::now() - startTime;
    }
    elapsedTime /= RUN_COUNT;
    timingResults.push_back(BenchmarkResult("ClassSingleBase to Level 3 [rttr_cast]         ", elapsedTime));
    printStats(timingResults);
}

/////////////////////////////////////////////////////////////////////////////////////////

BOOST_AUTO_TEST_CASE(testSingleClassHierachyEnd_PerformanceTest)
{
    using namespace boost::chrono;
    std::vector<BenchmarkResult> timingResults;
    std::vector<boost::shared_ptr<ClassSingleBase> > shared_vec;

    shared_vec.reserve(ITEM_COUNT * 5);
    for (int i = 0; i < ITEM_COUNT; ++i)
    {
        shared_vec.push_back(boost::make_shared<ClassSingle6A>());
        shared_vec.push_back(boost::make_shared<ClassSingle6B>());
        shared_vec.push_back(boost::make_shared<ClassSingle6C>());
        shared_vec.push_back(boost::make_shared<ClassSingle6D>());
        shared_vec.push_back(boost::make_shared<ClassSingle6E>());
    }
    std::vector<ClassSingleBase*> vec = transform(shared_vec);

    process_real_cpu_clock::time_point startTime;
    process_real_cpu_clock::duration elapsedTime;

    std::vector<int> result;
    result.resize(ITEM_COUNT * 5);
    for (int count = 0; count < RUN_COUNT; ++count)
    {
        startTime = process_real_cpu_clock::now();
        for (int i = 0; i < ITEM_COUNT * 5; ++i)
        {
            if (ClassSingle6A* item = dynamic_cast<ClassSingle6A*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle6B* item = dynamic_cast<ClassSingle6B*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle6C* item = dynamic_cast<ClassSingle6C*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle6D* item = dynamic_cast<ClassSingle6D*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle6E* item = dynamic_cast<ClassSingle6E*>(vec[i]))
            {
                result[i] = item->getType();
            }
        }
        elapsedTime += process_real_cpu_clock::now() - startTime;
    }
    elapsedTime /= RUN_COUNT;
    timingResults.push_back(BenchmarkResult("ClassSingleBase to Level 6 [dynamic_cast]      ", elapsedTime));

    elapsedTime = process_real_cpu_clock::duration();
    for (int count = 0; count < RUN_COUNT; ++count)
    {
        startTime = process_real_cpu_clock::now();

        for (int i = 0; i < ITEM_COUNT * 5; ++i)
        {
            if (ClassSingle6A* item = rttr_cast<ClassSingle6A*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle6B* item = rttr_cast<ClassSingle6B*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle6C* item = rttr_cast<ClassSingle6C*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle6D* item = rttr_cast<ClassSingle6D*>(vec[i]))
            {
                result[i] = item->getType();
            }
            else if (ClassSingle6E* item = rttr_cast<ClassSingle6E*>(vec[i]))
            {
                result[i] = item->getType();
            }
        }
        elapsedTime += process_real_cpu_clock::now() - startTime;
    }
    elapsedTime /= RUN_COUNT;
    timingResults.push_back(BenchmarkResult("ClassSingleBase to Level 6 [rttr_cast]         ", elapsedTime));
    printStats(timingResults);
}

/////////////////////////////////////////////////////////////////////////////////////////

BOOST_AUTO_TEST_CASE(testMultipleClassHierachy_PerformanceTest)
{
    using namespace boost::chrono;
    std::vector<BenchmarkResult> timingResults;
    std::vector<boost::shared_ptr<ClassMultipleBaseA> > shared_vec;

    shared_vec.reserve(ITEM_COUNT * 5);
    for (int i = 0; i < ITEM_COUNT; ++i)
    {
        shared_vec.push_back(boost::make_shared<FinalClass>());
        shared_vec.push_back(boost::make_shared<ClassMultiple1A>());
        shared_vec.push_back(boost::make_shared<ClassMultiple2A>());
        shared_vec.push_back(boost::make_shared<ClassMultiple3A>());
        shared_vec.push_back(boost::make_shared<ClassMultiple4A>());
        shared_vec.push_back(boost::make_shared<ClassMultiple5A>());
        shared_vec.push_back(boost::make_shared<ClassMultiple6A>());
    }
    std::vector<ClassMultipleBaseA*> vec = transform(shared_vec);

    process_real_cpu_clock::time_point startTime;
    process_real_cpu_clock::duration elapsedTime;

    std::vector<int> result;
    result.resize(ITEM_COUNT * 5);
    for (int count = 0; count < RUN_COUNT; ++count)
    {
        startTime = process_real_cpu_clock::now();
        for (int i = 0; i < ITEM_COUNT * 5; ++i)
        {
            if (FinalClass* item = dynamic_cast<FinalClass*>(vec[i]))
            {
                result[i] = item->getType();
            }
        }
        elapsedTime += process_real_cpu_clock::now() - startTime;
    }
    elapsedTime /= RUN_COUNT;
    timingResults.push_back(BenchmarkResult("ClassMultipleBaseA to FinalClass [dynamic_cast]", elapsedTime));

    elapsedTime = process_real_cpu_clock::duration();
    for (int count = 0; count < RUN_COUNT; ++count)
    {
        startTime = process_real_cpu_clock::now();

        for (int i = 0; i < ITEM_COUNT * 5; ++i)
        {
            if (FinalClass* item = rttr_cast<FinalClass*>(vec[i]))
            {
                result[i] = item->getType();
            }
        }
        elapsedTime += process_real_cpu_clock::now() - startTime;
    }
    elapsedTime /= RUN_COUNT;
    timingResults.push_back(BenchmarkResult("ClassMultipleBaseA to FinalClass [rttr_cast]   ", elapsedTime));
    printStats(timingResults);
}

/////////////////////////////////////////////////////////////////////////////////////////


BOOST_AUTO_TEST_CASE(testMultipleClassHierachy2_PerformanceTest)
{
    using namespace boost::chrono;
    std::vector<BenchmarkResult> timingResults;
    std::vector<boost::shared_ptr<ClassMulti1A> > shared_vec;

    shared_vec.reserve(ITEM_COUNT * 5);
    for (int i = 0; i < ITEM_COUNT; ++i)
    {
        shared_vec.push_back(boost::make_shared<ClassMulti7A>());
        shared_vec.push_back(boost::make_shared<ClassMulti6A>());
        shared_vec.push_back(boost::make_shared<ClassMulti5A>());
        shared_vec.push_back(boost::make_shared<ClassMulti4A>());
        shared_vec.push_back(boost::make_shared<ClassMulti3A>());
        shared_vec.push_back(boost::make_shared<ClassMulti2A>());
        shared_vec.push_back(boost::make_shared<ClassMulti1A>());
    }
    std::vector<ClassMulti1A*> vec = transform(shared_vec);

    process_real_cpu_clock::time_point startTime;
    process_real_cpu_clock::duration elapsedTime;

    std::vector<int> result;
    result.resize(ITEM_COUNT * 5);
    for (int count = 0; count < RUN_COUNT; ++count)
    {
        startTime = process_real_cpu_clock::now();
        for (int i = 0; i < ITEM_COUNT * 5; ++i)
        {
            if (ClassMulti4A* item = dynamic_cast<ClassMulti4A*>(vec[i]))
            {
                result[i] = item->getType();
            }
        }
        elapsedTime += process_real_cpu_clock::now() - startTime;
    }
    elapsedTime /= RUN_COUNT;
    timingResults.push_back(BenchmarkResult("ClassMulti1A to ClassMulti4A [dynamic_cast]    ", elapsedTime));

    elapsedTime = process_real_cpu_clock::duration();
    for (int count = 0; count < RUN_COUNT; ++count)
    {
        startTime = process_real_cpu_clock::now();

        for (int i = 0; i < ITEM_COUNT * 5; ++i)
        {
            if (ClassMulti4A* item = rttr_cast<ClassMulti4A*>(vec[i]))
            {
                result[i] = item->getType();
            }
        }
        elapsedTime += process_real_cpu_clock::now() - startTime;
    }
    elapsedTime /= RUN_COUNT;
    timingResults.push_back(BenchmarkResult("ClassMulti1A to ClassMulti4A [rttr_cast]       ", elapsedTime));
    printStats(timingResults);
}

/////////////////////////////////////////////////////////////////////////////////////////

BOOST_AUTO_TEST_CASE(testSingleClassHierachyEndTypeId_PerformanceTest)
{
    using namespace boost::chrono;
    std::vector<BenchmarkResult> timingResults;
    std::vector<boost::shared_ptr<ClassSingleBase> > shared_vec;

    shared_vec.reserve(ITEM_COUNT * 5);
    for (int i = 0; i < ITEM_COUNT; ++i)
    {
        shared_vec.push_back(boost::make_shared<ClassSingle6A>());
        shared_vec.push_back(boost::make_shared<ClassSingle6B>());
        shared_vec.push_back(boost::make_shared<ClassSingle6C>());
        shared_vec.push_back(boost::make_shared<ClassSingle6D>());
        shared_vec.push_back(boost::make_shared<ClassSingle6E>());
    }
    std::vector<ClassSingleBase*> vec = transform(shared_vec);

    process_real_cpu_clock::time_point startTime;
    process_real_cpu_clock::duration elapsedTime;

    std::vector<int> result;
    result.resize(ITEM_COUNT * 5);
    for (int count = 0; count < RUN_COUNT; ++count)
    {
        startTime = process_real_cpu_clock::now();
        for (int i = 0; i < ITEM_COUNT * 5; ++i)
        {
            if (typeid(ClassSingle6A) == typeid(*vec[i]))
            {
                result[i] = static_cast<ClassSingle6A*>(vec[i])->getType();
            }
            else if (typeid(ClassSingle6B) == typeid(*vec[i]))
            {
                result[i] = static_cast<ClassSingle6B*>(vec[i])->getType();
            }
            else if (typeid(ClassSingle6C) == typeid(*vec[i]))
            {
                result[i] = static_cast<ClassSingle6C*>(vec[i])->getType();
            }
            else if (typeid(ClassSingle6D) == typeid(*vec[i]))
            {
                result[i] = static_cast<ClassSingle6D*>(vec[i])->getType();
            }
            else if (typeid(ClassSingle6E) == typeid(*vec[i]))
            {
                result[i] = static_cast<ClassSingle6E*>(vec[i])->getType();
            }
        }
        elapsedTime += process_real_cpu_clock::now() - startTime;
    }
    elapsedTime /= RUN_COUNT;
    timingResults.push_back(BenchmarkResult("ClassSingleBase to Level 6 [typeid]            ", elapsedTime));

    elapsedTime = process_real_cpu_clock::duration();
    for (int count = 0; count < RUN_COUNT; ++count)
    {
        startTime = process_real_cpu_clock::now();

        for (int i = 0; i < ITEM_COUNT * 5; ++i)
        {
            if (TypeInfo::get<ClassSingle6A>() == TypeInfo::get(*vec[i]))
            {
                result[i] = static_cast<ClassSingle6A*>(vec[i])->getType();
            }
            else if (TypeInfo::get<ClassSingle6B>() == TypeInfo::get(*vec[i]))
            {
                result[i] = static_cast<ClassSingle6B*>(vec[i])->getType();
            }
            else if (TypeInfo::get<ClassSingle6C>() == TypeInfo::get(*vec[i]))
            {
                result[i] = static_cast<ClassSingle6C*>(vec[i])->getType();
            }
            else if (TypeInfo::get<ClassSingle6D>() == TypeInfo::get(*vec[i]))
            {
                result[i] = static_cast<ClassSingle6D*>(vec[i])->getType();
            }
            else if (TypeInfo::get<ClassSingle6E>() == TypeInfo::get(*vec[i]))
            {
                result[i] = static_cast<ClassSingle6E*>(vec[i])->getType();
            }
        }
        elapsedTime += process_real_cpu_clock::now() - startTime;
    }
    elapsedTime /= RUN_COUNT;
    timingResults.push_back(BenchmarkResult("ClassSingleBase to Level 6 [TypeInfo::get]     ", elapsedTime));
    printStats(timingResults);
}

BOOST_AUTO_TEST_SUITE_END()
